#!/usr/bin/env python
from __future__ import with_statement, division
import re
from innie import InnieParser
import nespalette
import sys
from firstfit import ffd_add, slices_union, slices_find, slices_remove

trace = False

# Parsing the config file ###########################################

class RomsetParser(InnieParser):
    """

.pages is a list whose elements are (name, list of title entries)
.pages_by_name is a dict from a page name to an index in .pages

"""
    
    def __init__(self, data=None, filenames=None):
        InnieParser.__init__(self)
        self.pages = []
        self.pages_by_name = {}
        self.cur_page = None
        self.addfilter(self.do_romset_pair)
        if data:
            self.readstring(data)
        if filenames:
            ok = self.read(filenames)

    def set_menu_page(self, name):
        if name not in self.pages_by_name:
            self.pages.append((name, []))
            self.pages_by_name[name] = len(self.pages) - 1
        self.cur_page = self.pages_by_name[name]

    def do_romset_pair(self, k, v):
        if k in ('#', ';'):
            return
        if k == 'page':
            self.set_menu_page(v)
            return
        if k == 'title':
            if len(self.pages[self.cur_page][1]) >= 20:
                self.set_menu_page('More')
            if len(self.pages[self.cur_page][1]) >= 20:
                raise IndexError('each page can have only 20 titles')
            self.pages[self.cur_page][1].append({})
        this_title = self.pages[self.cur_page][1][-1]
        this_title[k] = v

def get_entrypoint(rom):
    prg = rom['prg']
    pts = ((ord(prg[i-3]), ord(prg[i-4]), ord(prg[i-5]), ord(prg[i-6]))
           for i in range(0, len(prg), min(32768, len(prg))))
    pts = [(rh<<8|rl, nh<<8|nl)
           for (rh, rl, nh, nl) in pts]
    rom['resetpoints'] = [pt[0] for pt in pts]
    rom['nmipoints'] = [pt[1] for pt in pts]
    rom['base'] = [0x8000 for pt in pts]
    rom['patchpoints'] = [0xFFF0 for pt in pts]
    if len(prg) == 16384 and rom['resetpoints'][0] >= 0xC000:
        rom['base'] = [0xC000]
        rom['patchpoints'] = [0xFFF0]

hexrangeRE = re.compile(r'([0-9a-fA-F]+)-([0-9a-fA-F]+)')
def parse_prgunused(s):
    if not s:
        return None
    s = str(s).strip()
    m = hexrangeRE.match(s)
    if not m:
        raise ValueError("range '%s' must be two 4-digit base 16 ints separated by a hyphen"
                         % s)

    unusedStart = int(m.group(1), 16)
    unusedEnd = int(m.group(2), 16)
    if (unusedStart < 0x8000
        or unusedEnd < unusedStart
        or 0xFFFF < unusedEnd):
        raise ValueError("range %4x-%4x must be increasing and in 8000-FFFF"
                         % (t['title'], t['prgbank']))
        return None

    # Second number of a Python slice refers to first USED byte after
    # the unused range
    return (unusedStart, unusedEnd + 1)

patchRE = re.compile(r'\s*([89A-Fa-f][0-9A-Fa-f]{3})\s*[=:]\s*((?:[0-9A-Fa-f]{2})+)\s*')
def parse_patch(s):
    """Parse a binary patch.

A few of Shiru's games start by clearing all nametables, but due to a
mistake in programming, they end up clearing too much and scribbling
over the pattern tables as well.  This has no effect on a CHR ROM
mapper, but it becomes apparent when the game is inserted into a
CHR RAM multicart.  So I've added a mechanism for the config file to
fix such bugs directly in the binary during the build process.

"""
    from binascii import a2b_hex

    if not s:
        return None
    s = str(s).strip()
    m = patchRE.match(s)
    if not m:
        raise ValueError("range '%s' must be a 4-digit base 16 int, a : or =, and a hex string"
                         % s)
    return (int(m.group(1), 16), a2b_hex(m.group(2)))

players_types = {
    '1': 0,  # 1 player
    '2': 1,  # 2 players ONLY
    '1-2': 2,  # 1 or 2 players
    '1-2 alt': 3,  # 1 or 2 players alternating
    '1-3': 4,  # 1 to 3 players
    '1-4': 5,  # 1 to 4 players
}

def load_page_roms(pages):
    """Filter titles on pages to only those whose ROM can be loaded.

Return a tuple (page_titles, titles, roms, roms_by_name, unused_by_name).
page_titles is a list of (page name, index of first title on next page)
titles is a list of dicts, which generally have items
'title': string, 'description': longer string, 'author': string,
'rom': filename
roms is a list of (filename, romdata) pairs, where each romdata
is a dict returned by load_ines(), with elements 'prg' and 'chr'
and arrays 'base' and 'resetpoints' with one element per 32k bank
roms_by_name is a dict from filenames to an index into roms
unused_by_name is a dict from filenames to lists of unused byte slices

"""
    from ines import load_ines
    from os.path import basename

    page_titles = []
    roms_by_name = {}
    roms = []
    titles = []
    unused_by_name = {}
    all_patches = []

    lines = []
    for (pagename, titles_on_page) in pages:
        lines.append("\n== %s ==" % pagename)
        for t in titles_on_page:
            t.setdefault('title'), basename(t['rom'])
            t.setdefault('prgbank', '0')
            t.setdefault('chrbank', '0')
            t.setdefault('players', '1')
            
            try:
                romdata = roms[roms_by_name[t['rom']]][1]
                lines.append("Reusing %s" % t['rom'])
            except KeyError:
                try:
                    romdata = load_ines(t['rom'])
                    get_entrypoint(romdata)
                except StandardError, e:
                    print >>sys.stderr, "%s: ROM '%s' could not be loaded" % (t['title'], t['rom'])
                    print >>sys.stderr, str(e)
                    continue
                else:
                    roms.append((t['rom'], romdata))
                    roms_by_name[t['rom']] = len(roms) - 1
                    unused_by_name[t['rom']] = [set() for i in romdata['base']]

            epoch_year = 1970  # no NES until 1983, but for UNIX's sake
            try:
                year = int(t['year']) - epoch_year
            except KeyError:
                print >>sys.stderr, ("%s: year of first publication required skipping title"
                                     % (t['title']))
                continue
            except ValueError:
                print >>sys.stderr, ("%s: year '%s' must be a 4-digit integer; skipping title"
                                     % (t['title'], t['year']))
                continue
            if year < 0 or year > 2099 - epoch_year:
                print >>sys.stderr, ("%s: year %s must be between 1970 and 2099; skipping title"
                                     % (t['title'], epoch_year + year))
                continue
            t['year'] = year

            players = t['players'].replace(',', '-')
            try:
                players = players_types[players]
            except KeyError:
                print >>sys.stderr, ("%s: invalid number of players '%s'; skipping title"
                                     % (t['players']))
                continue
            t['players'] = players

            try:
                prgbank = int(t['prgbank'])
            except ValueError:
                print >>sys.stderr, ("%s: PRG bank '%s' must be an integer; skipping title"
                                     % (t['title'], t['prgbank']))
                continue
            if prgbank >= len(romdata['base']):
                print >>sys.stderr, ("%s: PRG bank %d must be less than %d, the number of banks in %s; skipping title"
                                     % (t['title'], prgbank, len(romdata['base']), t['rom']))
                continue
            t['prgbank'] = prgbank

            try:
                chrbank = int(t['chrbank'])
            except ValueError:
                print >>sys.stderr, ("%s: PRG bank '%s' must be an integer; skipping title"
                                     % (t['title'], t['prgbank']))
                continue
            if 'chr' not in romdata:
                chrbank = -1
            elif chrbank >= len(romdata['chr']) // 8192:
                print >>sys.stdout, ("%s: CHR bank %d must be less than %d, the number of banks in %s; skipping title"
                                     % (t['title'], chrbank, len(romdata['chr'] // 8192), t['rom']))
                continue
            t['chrbank'] = chrbank

            try:
                entrypoint = int(t['entrypoint'].lstrip('$'), 16)
            except KeyError:
                entrypoint = romdata['resetpoints'][prgbank]
            except ValueError:
                print >>sys.stderr, ("%s: entry point '%s' must be a hexadecimal integer; skipping title"
                                     % (t['title'], t['entrypoint']))
                continue
            if entrypoint < 0x8000 or entrypoint > 0xFFED:
                print >>sys.stderr, ("%s: entry point $%s must be in $8000-$FFED; skipping title"
                                     % (t['title'], t['entrypoint']))
                continue
            t['entrypoint'] = entrypoint
                                    

            prgunused = t.get('prgunused')
            if prgunused:
                okunused = {}
                for prgslice in prgunused.split(','):
                    try:
                        prgslice = parse_prgunused(prgslice)
                    except ValueError, e:
                        print >>sys.stderr, "%s: %s; ignoring" % (t['title'], e)
                    else:
                        unused_by_name[t['rom']][prgbank].add(prgslice)

            patches = t.get('patches')
            if patches:
                okunused = {}
                for patch in patches.split(','):
                    try:
                        patch = parse_patch(patch)
                    except ValueError, e:
                        print >>sys.stderr, "%s: %s; ignoring" % (t['title'], e)
                    else:
                        if patch:
                            all_patches.append((roms_by_name[t['rom']], prgbank)
                                               + patch)

            titles.append(t)
            # printable, or ignorable
            lines.extend(("", t['title']))
            t = sorted((k, v) for (k, v) in t.items() if k != 'title')
            lines.extend("  %s: %s" % row for row in t)
            lines.append("ROM size: %d KiB PRG + %d KiB CHR"
                         % (len(romdata['prg']) // 1024,
                            len(romdata.get('chr', '')) // 1024))
            t = sorted((k, v) for (k, v) in romdata.items()
                       if k not in ('prg', 'chr', 'trainer'))
            lines.extend("  %s: %s" % row for row in t)

        # Finalize the page
        if (len(titles) > 0
            and (len(page_titles) == 0 or len(titles) > page_titles[0][-1])):
            page_titles.append((pagename, len(titles)))

    if all_patches:
        lines.append("Patches: %s" % all_patches)
    if trace:
        print "\n".join(lines)
                               
    return (page_titles, titles, roms, roms_by_name, unused_by_name, all_patches)

def roms_to_banks(roms, unused_by_name):
    """Break up all ROMs into banks.

Return a 4-tuple (prgbanks, chrbanks, prg_starts, chr_starts).
prgbanks is [(array('B') of 32768 bytes, set of unused slices), ...]
chrbanks is just an array of 8-bit strings
prg_starts and chr_starts are arrays from indices into roms to
indices into prgbanks and chrbanks respectively
"""
    from array import array

    prgbanks = []
    chrbanks = []
    prg_starts = []
    chr_starts = []
    lines = []

    # At this point, split GNROMs into PRG banks, pad each NROM-128
    # to 32 KiB, and split CNROMs and GNROMs into CHR banks
    for (filename, romdata) in roms:
        prg_starts.append(len(prgbanks))
        chr_starts.append(len(chrbanks))

        # Get all PRG banks
        for i in range(len(romdata['base'])):
            unused_ranges = unused_by_name[filename][i]
            byte_offset = i * 0x8000
            prgdata = romdata['prg'][byte_offset:byte_offset + 0x8000]
            lines.append("%s bank %d: %d bytes" % (filename, i, len(prgdata)))
            prgbytes = array('B')

            # Pad $8000-$BFFF for high NROM-128
            prgbase = romdata['base'][i]
            prepadlen = prgbase - 0x8000
            if prepadlen > 0:
                unused_ranges.add((0x8000, prgbase))
                prgbytes.fromstring(array('B', [0xFF]) * prepadlen)
            prgbytes.fromstring(prgdata)

            # Pad $C000-$FFF9 and duplicate the vectors for
            # low NROM-128
            postpadlen = 0x7ffa - len(prgbytes)
            if postpadlen > 0:
                unused_ranges.add((0x8000 + len(prgbytes),
                                   min(0xfffa, 0x8000 + len(prgbytes) + postpadlen)))
                lines.append("Adding %d bytes of postpad" % postpadlen)
                prgbytes.fromstring(array('B', [0xFF]) * postpadlen)
                prgbytes.extend(prgbytes[0x3FFA:0x4000])

            if not unused_ranges:
                raise ValueError("%s: no prgunused range in bank %d" % (filename, i))

            unused_ranges = slices_union(unused_ranges)
            uur_txt = (', '.join("%x-%x" % (a, b-1)
                                 for (a, b) in unused_ranges))
            assert len(prgbytes) == 32768
            lines.append("PRG bank $%02x not using %s"
                         % (len(prgbanks), uur_txt))
            prgbanks.append((prgbytes, unused_ranges))

        if 'chr' in romdata:
            for byte_offset in range(0, len(romdata['chr']), 0x2000):
                chrdata = romdata['chr'][byte_offset:byte_offset + 0x2000]
                assert len(chrdata) == 8192
                lines.append("CHR bank $%02x"
                             % (len(chrbanks),))
                chrbanks.append(chrdata)

    if trace:
        print "\n".join(lines)
    return (prgbanks, chrbanks, prg_starts, chr_starts)

# The ROM starts with an unspecified bank loaded into $8000-$BFFF.
# To get into the menu, we have to patch each ROM bank's reset code
# to switch to the last bank.
#
# There are essentially two different patches.  The 10-byte patch can
# be placed only at $BFF0 or $FFF0.
# FFF0: sei ldx #$FF txs stx $FFF2 jmp ($FFFC)
# BFF0: sei ldx #$FF txs stx $BFF2 jmp ($FFFC)
# By the time the JMP operand is fetched, the mapper will have
# finished switching to the bank with the menu.
fff0_patch_data = "\x78\xA2\xFF\x9A\x8E\xF2\xFF\x6C\xFC\xFF"
bff0_patch_data = "\x78\xA2\xFF\x9A\x8E\xF2\xBF\x6C\xFC\xFF"

# The 20-byte patch can be placed anywhere in the ROM.
# x000: sei ldx #5
# x003: lda $x00E,x sta $F0,x dex bpl $x003 jmp $00F0
# x00E: sta $x013 jmp ($FFFC)
# Because of the 6502's lack of program-counter-relative addressing
# modes, the code needs to be "relocated", that is, a couple
# instructions need their addresses updated to reflect where the
# patch is inserted.
anywhere_patch_data = ("\x78\xA2\x05\xBD\x0E\x00\x95\xF0"
                       "\xCA\x10\xF8\x4C\xF0\x00\x8E\x13"
                       "\x00\x6C\xFC\xFF")
anywhere_patch_relocs = [4, 15]

def reloc_patch(patch_data, relocs, amount):
    """Relocate a piece of 6502 code to a given starting address."""
    patch_data = array.array('B', patch_data)
    for i in anywhere_patch_relocs:
        address = patch_data[i] + (patch_data[i + 1] << 8) + amount
        patch_data[i] = address & 0xFF
        patch_data[i + 1] = address >> 8
    return patch_data.tostring()

def find_best_patch_spot(rombankdata, unused_ranges):
    """Find the best spot for a reset patch in the unused ranges of a ROM bank.

It first tries to look for unused ranges at $FFF0-$FFF9 or
$BFF0-$BFF9.  If it can't find those, it'll look for 20 bytes at the
start or end of an unused range that are all the same value (e.g. all
$00 or $FF).  Otherwise, it uses the first range that's at least 20
bytes long.

Return a tuple of (start address, patched data string), or raise
IndexError if no good unused range is available.

"""

    # First prefer $FFF0 or $BFF0
    found_fff0 = slices_find(unused_ranges, (0xfff0, 0xfff0 + len(fff0_patch_data)))
    if found_fff0 >= 0:
        return (0xfff0, fff0_patch_data)
    found_bff0 = slices_find(unused_ranges, (0xbff0, 0xbff0 + len(fff0_patch_data)))
    if found_bff0 >= 0:
        return (0xbff0, bff0_patch_data)

    # Reject ranges that are too small to accept a relocatable patch
    unused_ranges = [(start, end)
                     for (start, end) in unused_ranges
                     if end - start >= len(anywhere_patch_data)]

    # If the start or end of an unused range is a run, use that
    for (start, end) in unused_ranges:
        for candidate in (start, end - len(anywhere_patch_data)):
            startoff = candidate - 0x8000
            endoff = startoff + len(anywhere_patch_data)
            if all(c == rombank[startoff]
                   for c in rombank[startoff + 1:endoff]):
                return (candidate,
                        reloc_patch(anywhere_patch_data,
                                    anywhere_patch_relocs,
                                    candidate))

    # Plop it in the first range.  Will raise IndexError if there are
    # no ranges.
    start = unused_ranges[0][0]
    return (start,
            reloc_patch(anywhere_patch_data, anywhere_patch_relocs, start))

def patch_bank(romdata, unused_ranges):
    """Add the reset patch to a bank at the best position.

Both romdata and unused_ranges are modified to reflect the patch.

Return the original reset vector and the data that the patch replaced.

"""
    from array import array

    (address, patch_data) = find_best_patch_spot(romdata, unused_ranges)

    # Patch the reset vector
    orig_reset = romdata[-4] | (romdata[-3] << 8)
    romdata[-4] = address & 0xFF
    romdata[-3] = address >> 8

    # Add the reset code
    offset = address - 0x8000
    unpatch_data = romdata[offset:offset + len(patch_data)].tostring()
    romdata[offset:offset + len(patch_data)] = array('B', patch_data)

    # Log the addition of the reset code.  We don't need to log the
    # address of the data that the reset patch replaced because
    # that's always equal to the NEW reset vector.
    slices_remove(unused_ranges, (address, address + len(patch_data)))
    return (orig_reset, unpatch_data)

def make_rom_directory(prg_starts, prg_lengths, unpatch_info, chr_starts, chr_total):
    """Turn an unpatch list into a ROM directory.

prg_starts is the index of the first PRG bank for each rom
prg_lengths is the length in bytes of each PRG ROM.  Can't infer this
from prg_starts because it doesn't distinguish NROM-128 (16384 bytes
before padding) from NROM-256 (32768 bytes and was not padded).
unpatch_info is a list of tuples, one for each bank:
(original reset vector, unpatch data)
chr_starts is the index of the first CHR bank for each ROM
chr_total is the total number of CHR banks in the ROM

Ordinarily, copylefted code and non-free code cannot be placed in the
same executable file.  But if the executable file is structured as an
archive, like a CD image, the "aggregate" exception of popular
copyleft licenses kicks in.  In order to qualify as an archive, the
multicart must allow lossless extraction of individual ROMs.

Extracting each page to a separate file is not enough because the
unpacker needs to do three things to be lossless:
1. combine the appropriate CHR bank or banks with the PRG data,
2. combine the PRG banks of a GNROM or BNROM multicart, and
3. reverse the patch to each PRG ROM bank that allows reset to find
   the menu no matter which bank is switched in on power-on.

The ROM directory gives instructions to turn the data in the
multicart's PRG ROM banks and compressed CHR segments back into
usable ROMs.  All data replaced by the reset patch is saved, except
in the case of NROM-128, where the reset patch is assumed to end up
in the padding.  Large ranges marked as unused in the config file are
replaced with other games' CHR ROM and screenshots, and the data that
they replace is not saved, but marking such large ranges is optional
anyway.  A collection of 53 NROM games, all patched at $BFF0
or $FFF0, should produce a ROM directory no bigger than 1 KiB.

Return a byte string with each ROM's data formatted as follows:

PRG size in 16 KiB units (1 byte) = np
CHR size in 8 KiB units (1 byte) = nc
PRG bank records (ceil(np/2) records of variable length)
CHR bank IDs (nc records of 1 byte each)

The format of a PRG bank record is as follows:
PRG bank index (1 byte)
Original contents of $FFFC-$FFFD, the reset vector (2 bytes)
Length of reset patch (1 byte; may be 0; bit 7 is 1 if the patch
covered a run of a single byte value)
Data that the reset patch replaced (1 byte if a run; otherwise as
long as the reset patch)

When writing the directory to the ROM, write a zero byte afterward so
that the extractor knows to stop extracting ROMs at a 0-length PRG.

When unpatching the ROM, replace data starting at the reset vector
with the data that the reset patch replaced, and then replace the
reset vector with the original one.

"""
    from array import array

    prg_starts = list(prg_starts)
    prg_starts.append(len(unpatch_info))
    chr_starts = list(chr_starts)
    chr_starts.append(chr_total)
    prg_lengths = [n // 16384 for n in prg_lengths]
    romdir = array('B')

    for i in range(len(prg_lengths)):
        prg_start_bank = prg_starts[i]
        prg_num_banks = prg_starts[i + 1] - prg_start_bank
        prg_num_halfbanks = prg_lengths[i]
        romdir.append(prg_num_halfbanks)
        chr_start_bank = chr_starts[i]
        chr_num_banks = chr_starts[i + 1] - chr_start_bank
        romdir.append(chr_num_banks)
        for j in range(prg_start_bank, prg_start_bank + prg_num_banks):
            (orig_fffc, unpatch_data) = unpatch_info[j]
            if prg_num_halfbanks < 2:  # don't unpatch NROM-128
                unpatch_data = ''
            unpatch_len = len(unpatch_data)

            # compress runs of unpatch data
            if (unpatch_len > 0
                and all(c == unpatch_data[0] for c in unpatch_data[1:])):
                unpatch_len = unpatch_len | 0x80
                unpatch_data = unpatch_data[0]
            romdir.extend([j, orig_fffc & 0xFF, orig_fffc >> 8, unpatch_len])
            romdir.fromstring(unpatch_data)
        romdir.extend(range(chr_start_bank, chr_start_bank + chr_num_banks))

    romdir.append(0)
    return romdir.tostring()

def create_blank_prg_bank():
    """Create a PRG bank that's all $FF except for a reset patch."""
    from array import array
    newbank = array('B', '\xff'*0x7FF0)
    newbank.fromstring(fff0_patch_data)
    if len(newbank) < 0x7FFA:
        newbank.fromstring('\x00'*(0x7FFA - fff0_patch_data))
    newbank.fromstring('\xF0\xFF\xF0\xFF\xF0\xFF')
    return newbank

def ffd_prg_factory():
    return (create_blank_prg_bank(), [(0x8000, 0xFFF0)])

def load_screenshot(filename):
    from PIL import Image
    from nespalette import convert_to_nes_colors
    from pilbmp2nes import pilbmp2chr
    
    imorig = Image.open(filename)
    (imnes, paldata) = convert_to_nes_colors(imorig, 0x0F)
    tiledata = pilbmp2chr(imnes)
    paldata = [c if c is not None else 0x0F for c in paldata]
    return (''.join(tiledata), paldata[1:4])

def load_screenshots(titles):
    """Load and compress all titles' screenshots.

This function takes a list of dictionaries with element 'screenshot',
loads the images using PIL, maps the screenshots to the NES color
palette, converts them to tiles, and compresses the tiles.

Return (screenshots, screenshots_by_titleno).
screenshots is [(pb53_bytes, [color1, color2, color3]), ...]
screenshot_ids is a list of one index into screenshots for each title

"""
    from pb53 import pb53

    screenshots = []
    screenshots_by_name = {}
    screenshot_ids = []
    for d in titles:
        filename = d.get('screenshot', '../tilesets/screenshots/default.png')
        try:
            scrid = screenshots_by_name[filename]
        except KeyError:
            (tiledata, pal) = load_screenshot(filename)
            scrid = len(screenshots)
            (ctiledata, midpoint) = pb53(tiledata)
            screenshots.append((ctiledata, pal))
            screenshots_by_name[filename] = scrid
        screenshot_ids.append(scrid)
    return (screenshots, screenshot_ids)

def insert_screenshots(titles, prgbanks):
    """Load screenshots and insert them into unused space.

Return a tuple (scrdir, screenshot_ids).
scrdir is a byte string with 6 bytes per entry:
bank, address low, address high, color 1, color 2, color 3
screenshot_ids is a list of scrdir entries, one for each title.
If i = screenshot_ids[titleno], then the title's entry is
scrdir[i * 6:i * 6 + 6].

"""
    # Load screenshots
    (screenshots, screenshot_ids) = load_screenshots(titles)

    # Insert screenshots into unused PRG ROM
    scr_sorted = sorted(enumerate(screenshots),
                        key=lambda (i, (d, paldata)): -len(d))
    scr_directory = [None for i in screenshots]
    for (i, (d, pal)) in scr_sorted:
        scr_directory[i] = ffd_add(prgbanks, d, ffd_prg_factory) + tuple(pal)
    if trace:
        print ("%d screenshots totaling %d compressed bytes plus %d for the directory"
               % (len(screenshots),
                  sum(len(d) for (d, paldata) in screenshots),
                  len(scr_directory)))

    # Format screenshot directory
    scrdir = ''.join(''.join((chr(b & 0xFF),
                              chr(a & 0xFF), chr(a >> 8),
                              chr(c1), chr(c2), chr(c3)))
                     for (b, a, c1, c2, c3) in scr_directory)
    return (scrdir, screenshot_ids)

def insert_chr(chrbanks, prgbanks):
    """Compress and insert the CHR banks into unused PRG ROM.

Return a byte string representing a directory of the compressed
CHR ROM, whose entries in the following format:

00: PRG bank
01-02: Address (little endian)
03-04: Midpoint (little endian)

The midpoint is used for decoding PB53 banks whose tiles in
$1000-$1FFF reference tiles in $0000-$0FFF.  The decoder runs two
instances of the pb53 decoder in parallel, and the second instance
copies tiles from the first.  The first starts from Address, the
second from (Address + Midpoint).

"""
    from pb53 import pb53

    total_unco = sum(len(c) for c in chrbanks)
    pb53banks = [pb53(c) for c in chrbanks]
    del chrbanks

    # Insert the CHR banks
    pb53banks_sorted = sorted(enumerate(pb53banks),
                              key=lambda (i, (d, mp)): -len(d))
    if False:
        print ("Compressed %d CHR bytes to %d"
               % (total_unco,
                  sum(len(c) + 4 for (c, sp) in pb53banks)))
        print "\n".join("CHR bank $%02x:%5d bytes" % (i, len(d))
                        for (i, (d, mp)) in pb53banks_sorted)
    chr_directory = [None for i in pb53banks]
    for (i, (d, mp)) in pb53banks_sorted:
        chr_directory[i] = ffd_add(prgbanks, d, ffd_prg_factory) + tuple(mp)
    chrdir = ''.join(''.join((chr(b & 0xFF),
                              chr(a & 0xFF), chr(a >> 8),
                              chr(mp & 0xFF), chr(mp >> 8)))
                     for (b, a, mp) in chr_directory)
    # At this point, chr_directory[] is a list of
    # (bank, address, midpoint) tuples, one for each CHR ROM bank
    # and chrdir is the NES-readable version of that

    if trace:
        print "CHR directory:"
        print "\n".join("CHR bank $%02x in PRG bank $%02x address $%02x"
                        % (i, b, a)
                        for (i, (b, a, m)) in enumerate(chr_directory))
    return chrdir

def pad_to_pow2m1(prgbanks):
    """Add blank PRG banks until reaching one less than a power of two."""

    # In binary, if (x & (x + 1)) is zero, then x is one less than
    # a power of two.
    while (len(prgbanks) & (len(prgbanks) + 1)) != 0:
        prgbanks.append((create_blank_prg_bank(), [(0x8000, 0xFFF0)]))

def make_title_directory(titles, roms_by_name,
                         prg_starts, chr_starts, screenshot_ids):
    """Make a machine-readable directory of ROM titles.

On the screen it is printed thus:

Title
(C) year Author
Number of players

Description (up to
16 lines)

So the title directory will have 16 bytes per entry:
PRG bank number
CHR bank number
Screenshot number
Year minus 1970
Number of players type
(3 bytes unused)
2-byte offset to title and author within the name block
2-byte offset to description within the description block
2-byte reset vector
(2 bytes unused)
For 64 entries this is only 1 KiB.

Each entry in the name and description blocks is terminated by a NUL
byte ('\0').  Each name block entry is a title and author separated
by a '\n', and the description block is likewise newline-delimited.

The title directory and the name block MUST be stored in the last
32 KiB bank of the ROM along with the menu.  Descriptions MAY
be in a separate bank, but they're all in one bank for a maximum
total of 32752 bytes.  In a 53-game collection, this limits the
descriptions to average a bit over half a kilobyte each, which is
fine considering 16 lines of VWF with about 28 characters per line.

Return a tuple of three 8-bit byte strings: the title directory,
the name block, and the description block.

"""
    from array import array

    # convert each title's PRG and CHR bank numbers from ROM-relative
    # to absolute
    prg_starts = [prg_starts[roms_by_name[t['rom']]] + t['prgbank']
                  for t in titles]
    chr_starts = [(chr_starts[roms_by_name[t['rom']]] + t['chrbank']
                   if t['chrbank'] >= 0
                   else 255)
                  for t in titles]

    titledir = array('B')
    name_block = array('B')
    desc_block = array('B')
    for (i, title) in enumerate(titles):
        reset = title['entrypoint']
        name_offset = len(name_block)
        desc_offset = len(desc_block)
        name_block.fromstring(title['title'])
        name_block.append(10)  # newline
        name_block.fromstring(title['author'])
        name_block.append(0)
        desc_block.fromstring(title['description'])
        desc_block.append(0)
        year = title['year']
        players = title['players']
        titledir_data = [
            prg_starts[i], chr_starts[i], screenshot_ids[i], year,
            int(players), 0, 0, 0,
            name_offset & 0xFF, name_offset >> 8,
            desc_offset & 0xFF, desc_offset >> 8,
            reset & 0xFF, reset >> 8,
            0, 0
        ]
        titledir.extend(titledir_data)

    return (titledir.tostring(), name_block.tostring(), desc_block.tostring())

def make_pagedir(pages):
    from array import array

    pageoffsets = array('B')
    pageoffsets.append(len(pages))
    pageoffsets.extend(p[1] for p in pages)
    for p in pages:
        pageoffsets.fromstring(p[0])
        pageoffsets.append(0)
    return pageoffsets.tostring()    

def main():
    from ines import load_ines
    from array import array

    # Load the config file
    pages = RomsetParser(filenames=["roms.cfg"]).pages

    # Load the ROMs
    (pages, titles, roms, roms_by_name, unused_by_name, all_patches) = load_page_roms(pages)
    if len(titles) == 0:
        raise IndexError("Not writing ROM: no titles were loaded")
    if trace:
        print ("%d titles across %d roms on %d pages loaded successfully"
               % (len(titles), len(roms), len(pages)))

    # Compute the length of each PRG ROM, so that the ROM directory
    # can distinguish NROM-128 from NROM-256
    prg_lengths = [len(romdata['prg']) for (nm, romdata) in roms]

    # Pack ROMs into the first PRG banks
    (prgbanks, chrbanks, prg_starts, chr_starts) \
               = roms_to_banks(roms, unused_by_name)
    del roms, unused_by_name

    if all_patches:
        for (i, prgbank, offset, data) in all_patches:
            prgbank = prgbanks[prg_starts[i] + prgbank][0]
            offset -= 0x8000
            prgbank[offset:offset + len(data)] = array('B', data)
        del prgbank, all_patches
        
    if trace:
        print "ROM size before adding CHR: %d PRG banks" % len(prgbanks)
    # Find the part of each PRG bank that will be replaced with a
    # patch, apply the patch, and save info needed to reverse
    # the patch
    unpatch_info = [patch_bank(romdata, unused_ranges)
                    for (romdata, unused_ranges) in prgbanks]
    # unpatch_info is one row for each bank:
    # [(original reset vector, unpatch address, unpatch data), ...]

    # Create the ROM directory
    romdir = make_rom_directory(prg_starts, prg_lengths, unpatch_info,
                                chr_starts, len(chrbanks))

    # Insert tile data for CHR ROM and screenshots
    chrdir = insert_chr(chrbanks, prgbanks)
    del chrbanks
    (scrdir, screenshot_ids) = insert_screenshots(titles, prgbanks)
    if trace:
        print "ROM size after inserting CHR and screenshots: %d PRG banks" % len(prgbanks)
    pad_to_pow2m1(prgbanks)
    if trace:
        print "Padded to power of 2: %d" % len(prgbanks)

    # Create the title directory
    (titledir, name_block, desc_block) \
               = make_title_directory(titles, roms_by_name,
                                      prg_starts, chr_starts, screenshot_ids)
    pagedir = make_pagedir(pages)

    # Save the ROM directory, CHR directory, screenshot directory,
    # title directory, and name block to the menu bank
    dirs_len = len(romdir) + len(chrdir) + len(scrdir) + len(titledir) + len(pagedir) + len(name_block)
    if trace:
        print ("All main-bank directories total %d bytes; description is %d"
               % (dirs_len, len(desc_block)))
    if dirs_len > 0x3FF0:
        raise ValueError("Directories are bigger than 16 KiB. this shouldn't happen.")

    final_bank = array('B', load_ines('../a53menu.nes')['prg'])
    final_banks = [(final_bank, [(0x8000, 0xBFF0)])]
    name_block_addr = ffd_add(final_banks, name_block)
    titledir_addr = ffd_add(final_banks, titledir)
    romdir_addr = ffd_add(final_banks, romdir)
    scrdir_addr = ffd_add(final_banks, scrdir)
    chrdir_addr = ffd_add(final_banks, chrdir)
    pagedir_addr = ffd_add(final_banks, pagedir)
    if trace:
        print "Remaining space in last bank:", final_banks[0][1]
    prgbanks.extend(final_banks)
    desc_block_addr = ffd_add(prgbanks, desc_block)

    # And finally make the key block at FF00
    # Format: ROM dir address, CHR dir address, screenshot dir address,
    # title dir address, page dir address, name block address,
    # desc block address, desc block bank, 0
    ff00 = array('B')
    ff00.extend([romdir_addr[1] & 0xFF, romdir_addr[1] >> 8,
                 chrdir_addr[1] & 0xFF, chrdir_addr[1] >> 8,
                 scrdir_addr[1] & 0xFF, scrdir_addr[1] >> 8,
                 titledir_addr[1] & 0xFF, titledir_addr[1] >> 8,
                 pagedir_addr[1] & 0xFF, pagedir_addr[1] >> 8,
                 name_block_addr[1] & 0xFF, name_block_addr[1] >> 8,
                 desc_block_addr[1] & 0xFF, desc_block_addr[1] >> 8,
                 desc_block_addr[0], 0])
    if trace:
        print "desc_block_addr is", desc_block_addr
    final_bank[0x7F00:0x7F10] = ff00

    iNESheader = array('B', "NES\x1A")
    iNESheader.append(len(prgbanks) * 2)
    iNESheader.append(0)  # CHR RAM
    mapperNumber = 34
    iNESheader.append(((mapperNumber & 0x0F) << 4) | 0x01)
    iNESheader.append(mapperNumber & 0xF0)
    iNESheader.extend(0 for i in range(8))

    with open("../a53games.nes", "wb") as outfp:
        outfp.write(iNESheader)
        outfp.writelines(b[0].tostring() for b in prgbanks)
    
if __name__ == '__main__':
    main()
